/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_FRAME_DESCRIPTOR_HH_
# define PCAP_FRAME_DESCRIPTOR_HH_

extern "C"
{
# include <inttypes.h>
}
# include <iostream>

# include <boost/shared_ptr.hpp>
# include <boost/shared_array.hpp>

# include <wipal/tool/static_base.hh>
# include <wipal/pcap/simple_types.hh>

namespace pcapxx
{

  // Forwad declarations.
  template <class>
  class descriptor;

  /**
   * Describe a PCAP frame/packet.
   *
   * A frame %descriptor holds a frame's @link pkthdr PCAP header
   * @endlink and bytes.
   *
   * Upon construction, both the frame's PCAP header and bytes are
   * deep-copied into the %descriptor.  You may therefore work with the
   * frame %descriptor after the original frame information has been
   * invalidated.
   *
   * Frame descriptors use shared pointers to store frame information.
   * You may copy them efficiently.  You may also retrieve smart pointer
   * to both the frame's PCAP header and bytes.
   */
  template <class PCAPDescriptor = tool::undefined_type>
  struct frame_descriptor
  {
    /// Smart pointer to a frame's PCAP header.
    typedef boost::shared_ptr<const pkthdr>	pkthdr_ptr;

    /// Smart pointer to a frame's bytes.
    typedef boost::shared_array<uint8_t>	bytes_ptr;

    /**
     * Construct a PCAP frame %descriptor from shared pointers to
     * @link PCAP headers @endlink and bytes.
     *
     * Since this constructor uses shared pointers, it does not
     * perform any deep copy.
     */
    frame_descriptor(const pkthdr_ptr& header, const bytes_ptr& bytes);

    /**
     * Build a PCAP frame %descriptor from an input stream.
     *
     * The stream position must be at the begining of a PCAP packet
     * header.
     *
     * @param[in]	desc	The descriptor the stream comes from.
     * @param[in]	stream	Input stream.
     * @param[in]	idx	Packet index.
     * @param[out]	pos	If not null, @c *pos is incremented with
     *				the number of read bytes.
     */
    template <class D>
    frame_descriptor(const descriptor<D>&	desc,
		     std::istream&		stream,
		     unsigned			idx,
		     std::streampos*		pos);

    /*
     * Build a frame %descriptor from another %descriptor which PCAP
     * descriptor has a different type.
     *
     * The resulting %descriptor has no refering PCAP descriptor.
     */
    template <class OtherDescriptor>
    frame_descriptor(const frame_descriptor<OtherDescriptor>&);

    /// Get the frame index inside the trace.
    unsigned		id() const;

    /// Get whether this frame come from an endianness-swapped trace.
    bool		swapped() const;

    /// Get a pointer to the frame's PCAP descriptor.  Might return 0.
    const PCAPDescriptor*
    desc_ptr() const;

    /**
     * Get the frame's PCAP descriptor.  Throws if frame has no such
     * descriptor.
     */
    const PCAPDescriptor&
    desc() const;

    /// Get a smart pointer to the frame's PCAP header.
    const pkthdr_ptr&	pcap_header() const;

    /// Get a smart pointer to the frame's data.
    const bytes_ptr&	bytes() const;

    /// Dump a frame %descriptor to a C++ stream.
    std::ostream&	print(std::ostream&) const;

    /// @name Equality checks
    //@{
    bool	operator == (const frame_descriptor&) const;
    bool	operator != (const frame_descriptor&) const;
    //@}

    /// Data embedded in a @link frame_descriptor frame %descriptor @endlink.
    struct shared_data
    {
      shared_data(const PCAPDescriptor*	desc,
		  unsigned		id,
		  const pkthdr_ptr&	pcap_header,
		  const bytes_ptr&	bytes);

      const PCAPDescriptor*	desc;	///<< Originating PCAP %descriptor.
      unsigned			id;	///<< Frame index inside the trace.
      pkthdr_ptr		pcap_header;	///<< PCAP header.
      bytes_ptr			bytes;		///<< Frame bytes.
    };

    /**
     * Return the object's shared data.
     *
     * Designed for conversions between @c frame_descriptor<T1> and @c
     * frame_descriptor<T2> (in frame_descriptor's constructors).  Do
     * not rely on any field in the returned object.  Please do not
     * use.
     */
    const boost::shared_ptr<shared_data>&	data() const;

  protected:
    boost::shared_ptr<shared_data>		data_;
  };

  /// Dump a @link frame_descriptor frame %descriptor @endlink to a C++ stream.
  template <class PCAPDescriptor>
  std::ostream&
  operator << (std::ostream&, const frame_descriptor<PCAPDescriptor>&);

} // End of namespace pcapxx.

# include "frame_descriptor.hxx"

#endif // ! PCAP_FRAME_DESCRITPOR_HH_
