/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_ADDR_HH_
# define WIFI_ADDR_HH_

extern "C"
{
# include <inttypes.h>
}

# include <string>
# include <ostream>

/// IEEE 802.11 compounds and operations.
namespace wifi
{

  /// IEEE 802.11 address.
  struct addr
  {
    /// Construct a null address.
    addr();

    /**
     * Construct an address from a memory location.
     *
     * An alternative to using this constructor is to @c static_cast
     * the memory location:
     *
     * @code
     * void* p = // To some memory containing a 802.11 address.
     *
     * std::cout << *static_cast<addr*> (p) << std::endl;
     * @endcode
     */
    addr(const void* a);

    /**
     * Construct an address from a string representation.
     *
     * @e Example:
     *
     * @code
     * addr a (std::string ("00:16:ec:92:84:b8"));
     * @endcode
     */
    addr(const std::string&);

    /// @name Comparison operators
    //@{
    bool		operator < (const addr&) const;
    bool		operator == (const addr&) const;
    bool		operator != (const addr&) const;
    //@}

    /// @name Address' bytes access
    //@{
    uint8_t&		operator [] (unsigned);
    const uint8_t&	operator [] (unsigned) const;
    //@}

    /// Copy to a memory location.
    void		dump(void *) const;

    /// Get <c>00:00:00:00:00:00</c>.
    static const addr&	null();

    /// Get <c>ff:ff:ff:ff:ff:ff</c>.
    static const addr&	broadcast();

  private:
    uint8_t	addr_[6];
  };

  /// Dump an @link addr address @endlink in human readable notation.
  std::ostream&
  operator << (std::ostream& os, const addr& a);

  /// Make a human readable string from an @link addr address @endlink .
  std::string
  make_string(const addr& a);

} // End of namespace wifi.

# include "addr.hxx"

#endif // ! WIFI_ADDR_HH_
