/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include <boost/preprocessor/tuple/elem.hpp>
#include <boost/preprocessor/seq/for_each.hpp>

#include "config.hh"

#include <wipal/tool/options.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/phy/empty.hh>
#include <wipal/phy/prism_header.hh>
#include <wipal/phy/rtap_header.hh>

namespace
{

  pcapxx::descriptor<>::link_type
  check_linktype(const pcapxx::list_of_traces<>& l,
		 const pcapxx::list_of_traces<>& r,
		 bool				 warn)
  {
    typedef pcapxx::descriptor<>::link_type	link_type;

#define P pcapxx::descriptor<>::PRISM_HEADER

    const link_type lt = l.empty() ? P : l.get(0).linktype();
    const link_type rt = r.empty() ? P : r.get(0).linktype();

#undef P

    l.expect(lt);
    r.expect(rt);

    if (warn and lt != rt)
      std::cerr << "WARNING: " << l.file_name() << " and " << r.file_name()
		<< " have different link types (" << lt << " and " << rt
		<< "), using " << lt << '.' << std::endl;

    return lt;
  }

  template <class D>
  bool
  eq_default(const pcapxx::frame_descriptor<D>& l,
	     const pcapxx::frame_descriptor<D>& r)
  {
    return l == r;
  }

  template <class D>
  bool
  eq_bytes(const pcapxx::frame_descriptor<D>& l,
	   const pcapxx::frame_descriptor<D>& r)
  {
    return
      0 == memcmp(l.bytes().get(), r.bytes().get(),
		  std::min(l.pcap_header()->caplen, r.pcap_header()->caplen));
  }

  template <class P, class D>
  bool
  eq_80211(const pcapxx::frame_descriptor<D>& l,
	   const pcapxx::frame_descriptor<D>& r)
  {
    return P::eq_80211(l, r);
  }

  template <class P, class D>
  bool
  eq_time_and_80211(const pcapxx::frame_descriptor<D>& l,
		    const pcapxx::frame_descriptor<D>& r)
  {
    return P::eq_time_and_80211(l, r);
  }

  template <class I1, class I2, class Eq>
  bool
  cmp(I1& l, const I1& lend,
      I2& r, const I2& rend,
      Eq eq)
  {
    while (l != lend and r != rend and eq(*l, *r))
      {
	++l;
	++r;
      }
    return l == lend and r == rend;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  const tool::opt::options options =
    tool::opt::parse_args(argc, argv,
			  WP_OPT_QUIET
			  WP_OPT_802CMP
			  WP_OPT_BYTESCMP
			  WP_OPT_DEFAULTCMP
			  WP_OPT_TIMECMP,
			  2, 0);

  try
    {
      typedef pcapxx::list_of_traces<>		list_of_traces;
      typedef pcapxx::descriptor<>::link_type	link_type;
      typedef list_of_traces::const_iterator	iterator;
      typedef pcapxx::descriptor<>		descriptor;

      list_of_traces	lhs (options.input[0]);
      list_of_traces	rhs (options.input[1]);
      const link_type	type = check_linktype(lhs, rhs, options.print_errors);

      iterator		l = lhs.begin();
      iterator		r = rhs.begin();
      const iterator	lend = lhs.end();
      const iterator	rend = rhs.end();

      bool		eq (false);

      switch (options.cmp)
	{
	case tool::opt::options::cmp_default:
	  eq = cmp(l, lend, r, rend, eq_default<descriptor>);
	  break;

	case tool::opt::options::cmp_802:
	  switch (type)
	    {

#define WP_cmp_802_case(R, Unused, LT)					\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      eq = cmp(l, lend, r, rend,				\
		       eq_80211<BOOST_PP_TUPLE_ELEM(2, 1, LT), descriptor>); \
	      break;

	      BOOST_PP_SEQ_FOR_EACH(WP_cmp_802_case, ~, WP_LINKTYPES)

#undef WP_cmp_802_case

	    default:
	      tool::opt::unsupported_linktype(type);
	      break;
	    }
	  break;

	case tool::opt::options::cmp_bytes:
	  eq = cmp(l, lend, r, rend, eq_bytes<descriptor>);
	  break;

	case tool::opt::options::cmp_time:
	  switch (type)
	    {

#define WP_cmp_time_case(R, Unused, LT)					\
	    case pcapxx::descriptor<>::BOOST_PP_TUPLE_ELEM(2, 0, LT):	\
	      eq = cmp(l, lend, r, rend,				\
		       eq_time_and_80211<BOOST_PP_TUPLE_ELEM(2, 1, LT), \
					 descriptor>);			\
	      break;

	      BOOST_PP_SEQ_FOR_EACH(WP_cmp_time_case, ~, WP_LINKTYPES)

#undef WP_cmp_802_case

	    default:
	      tool::opt::unsupported_linktype(type);
	      break;
	    }
	  break;

	default:
	  throw std::logic_error ("Unknown logic error.  Report to "
				  PACKAGE_BUGREPORT);
	}

      if (options.print_summary)
	{
	  std::cout << argv[0] << ": "
		    << lhs.file_name() << " and " << rhs.file_name() << ' ';
	  switch (eq)
	    {
	    case true:
	      std::cout << "are equals (after endianness adjustments.)";
	      break;
	    case false:
	      std::cout << "differs, frame " << (l == lend ? *r: *l) << '.';
	      break;
	    }
	  std::cout << std::endl;
	}
      return eq ? EXIT_SUCCESS : EXIT_FAILURE;
    }
  catch (std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
